/******************************************************************************
* ECE532 Digital System Design Course Spring 2007
* Dance Dance Revolution on FPGA
* Jeffrey Puk, Dharmendra Gupta, Kenny Chan
* 
* The main software control program for drawing arrows using a loop,
* and for streaming music and handling user input using interrupts
******************************************************************************/

#include "xparameters.h"	// generated system parameters
#include "xbasic_types.h"	// Xilinx basic types for device drivers
#include "xac97_l.h"			// AC97 driver
#include "xintc.h" 			// layer 1 interrupt controller device driver
#include "xgpio.h"			// gpio driver
#include "xuartlite.h"		// uart driver
#include "xuartlite_l.h"	// uart driver (low level)

#include "xexception_l.h"	
#include "xio.h"

#include "svga.h"

#define printf xil_printf   // a smaller footprint printf
#define DDR_BASEADDR XPAR_DDR_256MB_32MX64_RANK1_ROW13_COL10_CL2_5_MEM0_BASEADDR

// push button values
#define ENTER_BUTTON 0x0f
#define LEFT_BUTTON 0x1d
#define RIGHT_BUTTON 0x1e
#define UP_BUTTON 0x17
#define DOWN_BUTTON 0x1b

// UART receive values
#define UART_UP 65
#define UART_DOWN 66
#define UART_RIGHT 67
#define UART_LEFT 68
#define UART_W 119
#define UART_A 97
#define UART_S 115
#define UART_D 100

// number of iterations to run the game
#define ITERATIONS 20000

//************************** Global Variables ******************************/
XIntc InterruptController;
XGpio buttons;
XGpio dipswitch;
XUartLite uart;

extern struct arrow_t ** list;	// list of arrows from svga.h
extern struct arrow_t ** tail;	// list of tails from svga.h

int score = 0;							// the score for the game
BOOL resetScreen = FALSE;			// flag to clear arrows from screen

//************************** Function Prototypes ******************************/

void write_audio_to_mem(void * InstancePtr);
void push_button_handler(void * InstancePtr);
void uart_handler(void * InstancePtr, Xuint32 bytes);
void beat_handler(void * InstancePtr);

//************************** AC97 functions ******************************/

#define RECORD_MAX 0x150
volatile Xuint32 *ReadAddr;
volatile Xuint32 *WriteAddr;

void setPlayVolume(Xuint32 value)
{
 XAC97_WriteReg(XPAR_AUDIO_CODEC_BASEADDR, AC97_MasterVol, value);
  XAC97_WriteReg(XPAR_AUDIO_CODEC_BASEADDR, AC97_AuxOutVol, value);
  XAC97_WriteReg(XPAR_AUDIO_CODEC_BASEADDR, AC97_MasterVolMono, value);
  XAC97_WriteReg(XPAR_AUDIO_CODEC_BASEADDR, AC97_PCBeepVol, value);
  XAC97_WriteReg(XPAR_AUDIO_CODEC_BASEADDR, AC97_PCMOutVol, value);
}

void setRecordVolume(Xuint32 value)
{
	XAC97_WriteReg(XPAR_AUDIO_CODEC_BASEADDR, AC97_MicVol, value);
	XAC97_WriteReg(XPAR_AUDIO_CODEC_BASEADDR, AC97_LineInVol, value);
}

//************************** SVGA Loop ******************************/
void runSVGALoop()
{
	int i;
	int switch_data;
	int* pChar_data_wrd=(int*) SVGA_OFFSET;
	
	// clear all lists
	list = (struct arrow_t **)malloc(sizeof(struct arrow_t *)*4);
	tail = (struct arrow_t **)malloc(sizeof(struct arrow_t *)*4);
	for(i=0;i<4;i++)
	{
		list[i] = NULL;
		tail[i] = NULL;
	}
	
	// the loop, the game only runs for a fixed amount of iterations
	for (i=0;i<ITERATIONS;i++)
	{
		// clear the screen, draw the top grey arrows and print the score
		clearScreen(pChar_data_wrd);
		drawGreyArrowsToScreen(pChar_data_wrd);
		printScore(pChar_data_wrd, score);
		
		// generate a new set of arrows, draw them on the screen and
		//	delete arrows that reached the top of the screen
		generateArrows();
		drawArrowsToScreen(pChar_data_wrd);
		checkAndDeleteTailArrow();
		
		// if push button interrupt sets resetScreen
		if (resetScreen) 
		{
			resetScreen = FALSE;
			
			// free all arrows from memory
			freeAllArrowsFromLists();
						
			// clear the screen, draw top grey arrows and print the score
			clearScreen(pChar_data_wrd);
			drawGreyArrowsToScreen(pChar_data_wrd);
			printScore(pChar_data_wrd, score);
			
			sleep(1);	// wait one second
		}
		
		// get dip switch speed setting
		switch_data = XGpio_mGetDataReg(XPAR_DIPSWS_4BIT_BASEADDR,1);
		switch_data = switch_data & 0x0f;
		
		// delay screen drawing
		usleep(20000 + switch_data*2000);
	}
	
	// clean up memory
	freeAllArrowsFromLists();
}

//************************** Main program ******************************/
int main(void) 
{
	Xuint32 i;
	Xuint32 sample;
	XStatus Status;
	char c;

	// DDR memory address to read and write AC97 samples
	ReadAddr = (Xuint32 *)DDR_BASEADDR; 
	WriteAddr = (Xuint32 *)DDR_BASEADDR+25600;	// delay

	// ------------------ Interrupt setup ---------------------
 	printf("Initializing interrupts...\r\n");

	// Initialize the interrupt controller driver so that it's ready to use,
   // using the device ID that is generated in xparameters.h
   Status = XIntc_Initialize(&InterruptController, XPAR_OPB_INTC_0_DEVICE_ID);
   if (Status != XST_SUCCESS)
 		printf("Interrupt controller initialization error\n\r");

   // setup AC97 interrupt
	Status = XIntc_Connect(&InterruptController, XPAR_OPB_INTC_0_AUDIO_CODEC_INTERRUPT_INTR,
                           (XInterruptHandler)write_audio_to_mem,	NULL);
	XIntc_Enable(&InterruptController, XPAR_OPB_INTC_0_AUDIO_CODEC_INTERRUPT_INTR);
	if (Status != XST_SUCCESS)
       printf("AC97 Initialization error\n\r");

	// setup Push button interrupt
	XGpio_Initialize(&buttons, XPAR_PUSHBUTTONS_5BIT_DEVICE_ID);
	XGpio_SetDataDirection(&buttons, 1, 0x1F);  //All inputs = 0x1F
	Status = XIntc_Connect(&InterruptController, XPAR_OPB_INTC_0_PUSHBUTTONS_5BIT_IP2INTC_IRPT_INTR,
                           (XInterruptHandler)push_button_handler,	NULL);
	if (Status != XST_SUCCESS)
       printf("GPIO (Push Buttons) Initialization error\n\r");									
	XIntc_Enable(&InterruptController, XPAR_OPB_INTC_0_PUSHBUTTONS_5BIT_IP2INTC_IRPT_INTR);
	XGpio_InterruptClear(&buttons, XPAR_PUSHBUTTONS_5BIT_IP2INTC_IRPT_MASK);
	XGpio_InterruptEnable(&buttons, XPAR_PUSHBUTTONS_5BIT_IP2INTC_IRPT_MASK);
	XGpio_InterruptGlobalEnable(&buttons);

	// setup RS232 interrupt
	Status = XUartLite_Initialize(&uart, XPAR_RS232_UART_1_DEVICE_ID);
	if (Status != XST_SUCCESS)
       printf("Uart Initialization error\n\r");
	XUartLite_DisableInterrupt(&uart);
	XUartLite_ResetFifos(&uart);	
	Status = XIntc_Connect(&InterruptController, XPAR_OPB_INTC_0_RS232_UART_1_INTERRUPT_INTR,
                           (XInterruptHandler)uart_handler,	&uart);	
	XIntc_Enable(&InterruptController, XPAR_OPB_INTC_0_RS232_UART_1_INTERRUPT_INTR);
	XUartLite_EnableInterrupt(&uart);

	// setup beat interrupt
	Status = XIntc_Connect(&InterruptController, XPAR_OPB_INTC_0_BEAT_0_BEAT_INTR,
                           (XInterruptHandler)beat_handler,	NULL);
	XIntc_Enable(&InterruptController, XPAR_OPB_INTC_0_BEAT_0_BEAT_INTR);
	if (Status != XST_SUCCESS)
       printf("beat interrupt Initialization error\n\r");

   // Start the interrupt controller so interrupts are enabled for all
   // devices that cause interrupts. Specify real mode so that the timer
   // counter can cause interrupts through the interrupt controller.
   Status = XIntc_Start(&InterruptController, XIN_REAL_MODE);
   if (Status != XST_SUCCESS)
		printf("Interrupt controller start error\n\r");
   
	// Initialize the PPC405 exception table
	XExc_Init();

	// Register the interrupt controller handler with the exception table
   XExc_RegisterHandler(XEXC_ID_NON_CRITICAL_INT,
                         (XExceptionHandler)XIntc_VoidInterruptHandler, &InterruptController);
	// Enable non-critical exceptions
	XExc_mEnableExceptions(XEXC_NON_CRITICAL);


	// ------------------ AC 97 setup ---------------------
	printf("Initializing AC97...\r\n");
	XAC97_HardReset(XPAR_AUDIO_CODEC_BASEADDR);
	XAC97_EnableInput(XPAR_AUDIO_CODEC_BASEADDR, AC97_LINE_INPUT);
   XAC97_WriteReg(XPAR_AUDIO_CODEC_BASEADDR, AC97_PCM_DAC_Rate, AC97_PCM_RATE_48000_HZ);

	XAC97_ClearFifos(XPAR_AUDIO_CODEC_BASEADDR);
	XAC97_mSetControl(XPAR_AUDIO_CODEC_BASEADDR, AC97_ENABLE_IN_FIFO_INTERRUPT);
	XAC97_AwaitCodecReady(XPAR_AUDIO_CODEC_BASEADDR);
	
	// ------------------ SVGA loop ---------------------
	printf("starting svga loop...\n\r");
	runSVGALoop();

	return 0;
}

/* ISR to records samples from AC97 to DDR memory and plays back the recorded samples
* @param[in] InstancePtr instance pointer for passing private object
*/
void write_audio_to_mem(void * InstancePtr)
{
	Xuint32 sample;
	int i=0;

	// record samples and store to DDR memory
	for (i=0; i<256; i++) 
	{
		sample = XAC97_ReadFifo(XPAR_AUDIO_CODEC_BASEADDR);
		*WriteAddr = sample;
		WriteAddr++;
	}

	// plays back samples from DDR memory
	for (i=0; i<256; i++) 
	{
		XAC97_WriteFifo(XPAR_AUDIO_CODEC_BASEADDR, *ReadAddr);
		ReadAddr++;
	}
}

/* ISR to handle push button presses
* @param[in] InstancePtr instance pointer for passing private object
*/
void push_button_handler(void * InstancePtr)
{
	const int slack = 10;	// amount of slack for user to hit the arrow at the top
	unsigned int pb_data;
	pb_data = XGpio_mGetDataReg(XPAR_PUSHBUTTONS_5BIT_BASEADDR,1);
	
	printf("kenny was here!\n\r");
	switch(pb_data)
	{
		// reset by removing arrows from the screen and resetting score
		case ENTER_BUTTON:
			resetScreen = TRUE;
			score=0;
		break;
		
		// check if top arrow is near the top when push button is pressed
		//  if so, remove top arrow and increment score by 100
		case LEFT_BUTTON:
			if(tail[0]->y < slack)	
			{
				tail[0]->y = -10;
				score = score + 100;
			}
		break;
		case UP_BUTTON:
			if(tail[1]->y < slack)	
			{
				tail[1]->y = -10;
				score = score + 100;
			}
		break;
		case DOWN_BUTTON:
			if(tail[2]->y < slack)
			{
				tail[2]->y = -10;
				score = score + 100;
			}
		break;
		case RIGHT_BUTTON:
			if(tail[3]->y < slack)	
			{
				tail[3]->y = -10;
				score = score + 100;
			}
	}

	XGpio_InterruptClear(&buttons, XPAR_PUSHBUTTONS_5BIT_IP2INTC_IRPT_MASK);
}

/* ISR to handle UART traffic
* @param[in] InstancePtr instance pointer for passing private object
*/
void uart_handler(void * InstancePtr, Xuint32 bytes)
{
   Xuint32 IsrStatus;
	Xuint8 c;
	
	// amount of slack for user to hit the arrow, counted from the top of grey arrows
	const int slack = 10;	

   // Read the status register to determine which, coulb be both interrupt is active
   IsrStatus = XIo_In32(uart.RegBaseAddress + XUL_STATUS_REG_OFFSET);

	// receive interrupt
   if ((IsrStatus & (XUL_SR_RX_FIFO_FULL | XUL_SR_RX_FIFO_VALID_DATA)) != 0) 
   {
		c = (Xuint8)XIo_In32(XPAR_RS232_UART_1_BASEADDR + XUL_RX_FIFO_OFFSET);
		switch (c)
		{
			// check if top arrow is near the top when push button is pressed
			//  if so, remove top arrow and increment score by 100
			//case UART_LEFT: 
			case UART_A:
				if(tail[LEFT]!=NULL && tail[LEFT]->y < slack)	
				{
					tail[LEFT]->y = -10;
					score = score + 100;
				}
			break;
			//case UART_UP: 
			case UART_W:
				if(tail[UP]!=NULL && tail[UP]->y < slack)	
				{
					tail[UP]->y = -10;
					score = score + 100;
				}
			break;
			//case UART_DOWN: 
			case UART_S:
				if(tail[DOWN]!=NULL && tail[DOWN]->y < slack)	
				{
					tail[DOWN]->y = -10;
					score = score + 100;
				}			
			break;
			//case UART_RIGHT: 
			case UART_D:
				if(tail[RIGHT]!=NULL && tail[RIGHT]->y < slack)	
				{
					tail[RIGHT]->y = -10;
					score = score + 100;
				}			
		}
   }
	// transmit interrupt
	if ((IsrStatus & XUL_SR_TX_FIFO_EMPTY) != 0)	
   {
		//printf("t ");
   }

	// clear uart fifo to reset interrupt
	XUartLite_ResetFifos(&uart);
}

/* ISR to handle interrupt coming from beat detection module
* @param[in] InstancePtr instance pointer for passing private object
*/
void beat_handler(void * InstancePtr)
{
	score++;	// test by incrementing score by 1
}
