//////////////////////////////////////////////////////////////////
// beat.v
//
// beat detection algorithm. For details on operation, please see README
// Rev 1.0a
// Date Mar 20, 2007
//
// beat.v
//  +multiplier (black box)
//  +accumulator
//  +a_lt_b
//  +mult15.v
//  +average.v
//   +fifon.v
/////////////////////////////////////////////////////////////////

module beat(clk, resetn, fft_data_re, fft_data_im, fft_index, fft_edone, beat);
    //parameter fft_data_size = 27; //amount of data
    //parameter fft_num_bit = 10; //1024pt fft = 2^10 (for index bits)
    input clk, resetn, fft_edone;
    input [26:0] fft_data_re, fft_data_im;
    input [9:0] fft_index;
    output beat;
    
    parameter S1=2'b00, S2=2'b01, S3=2'b10;
    reg [1:0] this_state, next_state;
    reg CAcc, EAvg;
    
    wire [51:0] re2, im2; // output of re^2 and im^2 (52 each, 53 sum)
    wire [57:0] Es; //energy^2 in each subband (sum of 32 numbers, max+5 bits) (53+5=58)
    wire [62:0] nEs, nAvg; // n x Es and n x Avg (for comparison)
    wire [63:0] nEs15; //nEs x 1.5 (coefficient for beat detection)
    
    always @ (posedge clk or negedge resetn) begin
        if (~resetn) begin
            this_state<=S1;
        end
        else begin
            this_state<=next_state;
        end
        
    end
    
    always @(this_state, fft_edone, fft_index) begin
        case(this_state)
            S1:   if (fft_edone) next_state=S2;
                  else           next_state=S1;
            S2:   if (fft_index==32) next_state=S3;
                  else           next_state=S2;
            S3:                  next_state=S1;
            default:             next_state=S1;
        endcase
    end
    
    always @(this_state) begin
        //default
        CAcc=0; EAvg=0;
        case(this_state)
            S1:   CAcc=1;
            S3:   EAvg=1;
        endcase
    end
    
    // power calculation
    multiplier MRe(fft_data_re[25:0], fft_data_re[25:0], re2);
    multiplier MIm(fft_data_im[25:0], fft_data_im[25:0], im2);

    // total power in subband of power
    accumulator Acc(53'b0+re2+im2, Es, clk, CAcc);

    // average of power    
    average Avg(clk, 1'b0, EAvg, Es, nEs, nAvg);
       defparam Avg.nbit=5;
       defparam Avg.n=32; //store 32 numbers
       defparam Avg.p=58; // each number is 58 bit, output = 58+5 = 63 bits

    // input x coefficient (get threshold)
    mult15n coeff(nEs, nEs15);
       defparam coeff.n=63;

    // compare with average
    a_lt_b compare(.ce(EAvg), .clk(clk), .qa_lt_b(beat), .a({1'b0, nAvg}), .b(nEs15)); //CE, CLK, A<B, A, B
endmodule
