%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% ECE532 Digital System Design Course Spring 2007
% Dance Dance Revolution on FPGA
% Jeffrey Puk, Dharmendra Gupta, Kenny Chan
%
% Matlab file to simulate beat detection algorithm 
%  (frequency selected sound energy method)
%   In this algorithm, we want to detect whether an instant in time has a
%   beat.  This is done by selecting a particular subband and measure the 
%
%   Reference: Beat detection Algorithm from YOV408 programming resources
%   http://www.yov408.com/html/articles.php?p=1
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% 

clear;     % clear variables from workspace

%--------------------------------------------------------------------------

[wav sr] = wavread('aquabutterfly.wav');    % read in the song
wav = wav(500000:2500000,1);                % read only a part of it
wav = wav(:,1);   % setting 1 channel and making a sample

%--------------------------------------------------------------------------

clipvalue = 6;      % the maximum value to graph
spi = 1024;         % number of samples per instant (usually 1024)

wavsize = size(wav);
numsamples = floor(wavsize(1) / spi);   % number of 1024 samples
subbandsize =32;                        % size of the subband (usually 32)
numsubbands = spi / subbandsize;        % number of subbands
numhistory = 64;%floor(sr / spi);       % takes approx 1 second of local history 
                                        %(changed to 64 for easier implementation)

E = ones(numsubbands, numhistory);      % array to store history of instants

C = zeros(1,numsubbands);               % ratio of current instant vs. history
maxC = zeros(1,numsamples);             % store current ratio (picked lowest subband)
maxCout = zeros(1,numsamples);          % store recent valid beat(validity based on moving average)

% variables for implementing moving average, 
%   a spike is when ratio met threshold value
lastSpike = 1;                          % last j sample when it spiked
lastSpikeTime = 0;                      % last time it spiked
avgSpikes = ones (1,50) .*10000;        % array to store moving average samples
spikeThreshold = 1.5;                   % threshold for a spike

%--------------------------------------------------------------------------
% for each instant sample
for (j=1:numsamples)
  
    % perform fft on a sample
    fftx = fft(wav( (spi*(j-1))+1 : (spi*(j-1))+spi ), spi);
    B = (abs(fftx).^2) ;

    % calculate the energy for each subband
    Es = zeros(subbandsize,1);
    for (i=0:numsubbands-1)
       Es(i+1) = sum(B(i*subbandsize+1:(i+1)*subbandsize)); % numsubbands/spi; %reduced calc
    end
    
    % calculate the mean of each subband
    meanE = zeros(numsubbands,1);
    for (i=1:numsubbands)
        meanE(i) = mean(E(i,:));
    end
        
    % shift out oldest sample and add new energy values
    E = circshift(E, [0 1]);
    E(1:numsubbands,1) = Es(1:numsubbands);
    
    % store the ratio of current instant over the mean for each subband
    for (i=1:numsubbands)
        if (meanE~=0)
            C(i) = Es(i)/meanE(i);
        else
            C(i)=0;
        end
    end
    
    % take only one subband information as the beat
    maxC(j)=C(1); 
    
    % test if current beat meets a certain threshold
    if (maxC(j) > spikeThreshold)
        lastSpikeTime = j - lastSpike;  % time since last spike
        lastSpike = j;                  % update last spike value to current sample
        
        avgSpikes = circshift(avgSpikes, [0 1]);    % shift value into moving average
        avgSpikes(1) = lastSpikeTime;
    end

    % get the new moving average
    meanAvgSpikes = mean(avgSpikes);
    
    % if current subband meets a certain threshold and the spike is some
    % time away from the last spike, then register current sample as a valid beat
    if (maxC(j) > spikeThreshold & lastSpikeTime > meanAvgSpikes)
        maxCout(j) = 2; % stem
    else
        maxCout(j) = 0; % no stem
    end
    
    % clip current beat value for displaying
    if (maxC(j)>clipvalue)
        maxC(j) = clipvalue;
    end
end    

%--------------------------------------------------------------------------

% plot all the beat values
figure;
plot(maxC);

%--------------------------------------------------------------------------

% plot the valid beat stems as the music plays
fig1 = figure;
viewSegment = 2000; % allocate x-axis to display a set number of samples
set(fig1,'Position',[10 1000 1910 400]);    % position the plot on screen

% draw the axis and set up variables for plotting
axes1 = axes('xlim',[0 viewSegment],'ylim',[0 clipvalue]) ;
x = [1:numsamples];
CDisplay = zeros(1,numsamples);
hline = line(x,CDisplay);

player=audioplayer(wav,sr);

% waittime = spi/sr;
% waittime = wavsize(1)/sr/(wavsize(1)/spi);
waittime = spi/sr*1.3;
spiCount = 0;   % audio-sample per iteration count

for (j=1:numsamples)
    % clear the stems if they run out of the graph
    if (mod(j,viewSegment)==0)
        clf;
        axes1 = axes('xlim',[j j+viewSegment],'ylim',[0 clipvalue]) ;
        hline = line(x,CDisplay);
    end
    
    % set data to be displayed
    CDisplay(j)=maxCout (j);

    % thicken line and set line color when there is a valid beat
    % if (CDisplay(j) > 1.5)
    %   set(hline,'Color','red');
    %   set(hline,'LineWidth',1.0);
    % else
    %   set(hline,'Color', 'blue');
    %   set(hline,'LineWidth',0.5);
    % end
    
    set(hline,'ydata',CDisplay);    % draw the new stem for this sample
    play(player,spiCount+1);        % play the current sample
    pause(waittime);                % pause loop execution for music to play at relatively correct speed
    stop(player);                   % stop playing

    spiCount = spiCount + spi;      % increment to next sample
end
